#include "HakoiriMusume.h"
#include <iostream>
#include <fstream>

using namespace std;

const int HakoiriMusume::OBJECT_SIZE_W[OBJECT_TYPE_NUM] = { 2,1,2,1 };
const int HakoiriMusume::OBJECT_SIZE_H[OBJECT_TYPE_NUM] = { 2,2,1,1 };

int HakoiriMusume::OBJECT_NUM;

//ID@S̈TYPE
//-1 
HakoiriMusume::HakoiriMusume(int Qnum) 
{
	char filename[64];
	sprintf(filename, "HakoiriMusumeQ%d.txt", Qnum);
	ifstream ifs(filename);
	if (!ifs.good()) { cout << "ԍ܂B" << endl;}
	ifs >> OBJECT_NUM;
	for (int i = 0; i < SIZE_Y; i++)
	{
		for (int j = 0; j < SIZE_X; j++)
		{
			ifs >> c[j][i];
		}
	}
	int count = 0;
	bool flag = true;
	for (int i = 0; i < SIZE_Y; i++)
	{
		for (int j = 0; j < SIZE_X; j++)
		{
			if (c[j][i] < 0) 
			{
				space_x[count] = j; space_y[count] = i;
				count++;
			}
			if (flag&&c[j][i] == 0) 
			{
				musume_x = j; musume_y = i;
				flag = false;
			}
		}
	}
	ifs.close();
}

HakoiriMusume::HakoiriMusume(const HakoiriMusume *p, const NodeBasePtr parent) :NodeBase(parent)
{
	for (int i = 0; i < SIZE_X; i++)
	{
		for (int j = 0; j <SIZE_Y; j++) 
		{
			c[i][j] = p->c[i][j];
		}
	}
	for (int i = 0; i < SPACE_NUM; i++) 
	{
		space_x[i] = p->space_x[i];
		space_y[i] = p->space_y[i];
	}
	musume_x = p->musume_x;
	musume_y = p->musume_y;
}

void HakoiriMusume::expand(std::list<NodeBasePtr>* nodeList, const NodeBasePtr parent) const
{
	bool** moved;
	moved = new bool*[OBJECT_NUM];
	for (int i = 0; i < OBJECT_NUM; i++)moved[i] = new bool[4];

	HakoiriMusume* expandNode;
	for (int i = 0; i < OBJECT_NUM; i++)
	{
		for (int j = 0; j < 4; j++)
		{
			moved[i][j] = false;
		}
	}
	for (int i = 0; i < SPACE_NUM; i++)
	{
		int sx = space_x[i], sy = space_y[i];
		if (sx > 0 && IS_OBJECT(c[sx - 1][sy]) && !moved[c[sx - 1][sy] % 100][RIGHT])
		{
			if (judgeMovable(sx - 1, sy, RIGHT))
			{
				moved[c[sx - 1][sy] % 100][RIGHT] = true;
				expandNode = new HakoiriMusume(this, parent);
				move(expandNode,sx-1, sy, RIGHT);
				nodeList->push_back(NodeBasePtr(expandNode));
			}
		}
		if (sy > 0 && IS_OBJECT(c[sx][sy - 1]) && !moved[c[sx][sy - 1] % 100][DOWN])
		{
			if (judgeMovable(sx, sy - 1, DOWN))
			{
				moved[c[sx][sy - 1] % 100][DOWN] = true;
				expandNode = new HakoiriMusume(this, parent);
				move(expandNode, sx, sy-1, DOWN);
				nodeList->push_back(NodeBasePtr(expandNode));
			}
		}
		if (sx + 1 < SIZE_X &&IS_OBJECT(c[sx + 1][sy]) && !moved[c[sx + 1][sy] % 100][LEFT])
		{
			if (judgeMovable(sx + 1, sy, LEFT))
			{
				moved[c[sx + 1][sy] % 100][LEFT] = true;
				expandNode = new HakoiriMusume(this, parent);
				move(expandNode, sx+1, sy, LEFT);
				nodeList->push_back(NodeBasePtr(expandNode));
			}
		}
		if (sy + 1 < SIZE_Y &&IS_OBJECT(c[sx][sy + 1]) && !moved[c[sx][sy + 1] % 100][UP])
		{
			if (judgeMovable(sx, sy + 1, UP))
			{
				moved[c[sx][sy + 1] % 100][UP] = true;
				expandNode = new HakoiriMusume(this, parent);
				move(expandNode, sx, sy+1, UP);
				nodeList->push_back(NodeBasePtr(expandNode));
			}
		}
	}
	if (nodeList->empty()) 
	{
		moved[0][0] = false;
	}
	for (int i = 0; i < OBJECT_NUM; i++)delete[] moved[i];
	delete[] moved;
}

bool HakoiriMusume::judgeMovable(int x_, int y_, eDirection direction) const
{
	if (IS_SPACE(c[x_][y_]))return false;
	int type = c[x_][y_]/100;
	int x = x_, y = y_;
	while (x > 0 && c[x - 1][y] == c[x][y])x--;
	while (y > 0 && c[x][y - 1] == c[x][y])y--;
	switch (direction)
	{
	case UP:
		if (y - 1 < 0)return false;
		for (int i = x; i < x + OBJECT_SIZE_W[type]; i++)
		{
			if (IS_OBJECT(c[i][y - 1]))return false;
		}
		break;
	case DOWN:
		if (y + OBJECT_SIZE_H[type] >= SIZE_Y)return false;
		for (int i = x; i < x + OBJECT_SIZE_W[type]; i++)
		{
			if (IS_OBJECT(c[i][y + OBJECT_SIZE_H[type]]))return false;
		}
		break;
	case LEFT:
		if (x - 1 < 0)return false;
		for (int i = y; i < y + OBJECT_SIZE_H[type]; i++)
		{
			if (IS_OBJECT(c[x - 1][i]))return false;
		}
		break;
	case RIGHT:
		if (x + OBJECT_SIZE_W[type] >= SIZE_X)return false;
		for (int i = y; i < y + OBJECT_SIZE_H[type]; i++)
		{
			if (IS_OBJECT(c[x + OBJECT_SIZE_W[type]][i]))return false;
		}
		break;
	}
	return true;
}

void HakoiriMusume::move(HakoiriMusume *node, int ox_, int oy_, eDirection direction)
{
	int type = node->c[ox_][oy_] / 100;
	if (node->c[ox_][oy_] == 0)
	{
		switch (direction)
		{
		case UP:node->musume_y--; break;
		case DOWN:node->musume_y++; break;
		case LEFT:node->musume_x--; break;
		case RIGHT:node->musume_x++; break;
		}
	}
	if (node->musume_y == 2) 
	{
		node->musume_x = node->musume_x + 1 - 1;
	}
	int ox = ox_, oy = oy_;
	while (ox > 0 && node->c[ox - 1][oy] == node->c[ox][oy])ox--;
	while (oy > 0 && node->c[ox][oy - 1] == node->c[ox][oy])oy--;
	int temp, spaceid;
	switch (direction)
	{
	case UP:
		for (int i = ox; i < ox + OBJECT_SIZE_W[type]; i++)
		{
			spaceid = SPACE_ID(node->c[i][oy - 1]);
			temp = node->c[i][oy + OBJECT_SIZE_H[type] - 1];
			node->c[i][oy + OBJECT_SIZE_H[type] - 1] = node->c[i][oy - 1];
			node->c[i][oy - 1] = temp;
			node->space_y[spaceid] += OBJECT_SIZE_H[type];			
		}
		break;
	case DOWN:
		for (int i = ox; i < ox + OBJECT_SIZE_W[type]; i++)
		{
			spaceid = SPACE_ID(node->c[i][oy + OBJECT_SIZE_H[type]]);
			temp = node->c[i][oy + OBJECT_SIZE_H[type]];
			node->c[i][oy + OBJECT_SIZE_H[type]] = node->c[i][oy];
			node->c[i][oy] = temp;
			node->space_y[spaceid] -= OBJECT_SIZE_H[type];
		}
		break;
	case LEFT:
		for (int i = oy; i < oy + OBJECT_SIZE_H[type]; i++)
		{
			spaceid = SPACE_ID(node->c[ox - 1][i]);
			temp = node->c[ox + OBJECT_SIZE_W[type] - 1][i];
			node->c[ox + OBJECT_SIZE_W[type] - 1][i] = node->c[ox - 1][i];
			node->c[ox - 1][i] = temp;
			node->space_x[spaceid] += OBJECT_SIZE_W[type];
		}
		break;
	case RIGHT:
		for (int i = oy; i < oy + OBJECT_SIZE_H[type]; i++)
		{
			spaceid = SPACE_ID(node->c[ox + OBJECT_SIZE_W[type]][i]);
			temp = node->c[ox + OBJECT_SIZE_W[type]][i];
			node->c[ox + OBJECT_SIZE_W[type]][i] = node->c[ox][i];
			node->c[ox][i] = temp;
			node->space_x[spaceid] -= OBJECT_SIZE_W[type];
		}
		break;
	}
}

bool HakoiriMusume::finished()const 
{
	return musume_x == GOAL_X&&musume_y == GOAL_Y;
}

void HakoiriMusume::print() const
{
	for (int i = 0; i < SIZE_X; i++)cout << "-";
	cout << endl;
	for (int y = 0; y < SIZE_Y; y++) 
	{
		for (int x = 0; x < SIZE_X; x++) 
		{
			if (IS_SPACE(c[x][y])) cout << "-";
			else
			{
				if (c[x][y] % 100 < 10)cout << c[x][y] % 100;
				else {
					char letter = c[x][y] % 100 + 97 - 10;
					cout << letter;
				}
			}
		}
		cout << endl;
	}
	
	for (int i = 0; i < SPACE_NUM; i++) 
	{
		cout << "sx[" << i << "]=" << space_x[i] << ",sy[" << i << "]=" << space_y[i] << endl;
	}
	cout << "mx="<< musume_x << ",my="<< musume_y << endl;
	
	for (int i = 0; i < SIZE_X; i++)cout << "-";
	cout << endl;
}

bool HakoiriMusume::equalTo(const NodeBasePtr ptr) const 
{
	shared_ptr<HakoiriMusume> a = dynamic_pointer_cast<HakoiriMusume>(ptr);
	bool equal1flag = true;
	for (int i = 0; i < SIZE_X; i++) 
	{
		for (int j = 0; j < SIZE_Y; j++) 
		{
			if (c[i][j] / 100 != a->c[i][j] / 100)
			{
				equal1flag = false;
				break;
			}
		}
		if (!equal1flag)break;
	}
	if (equal1flag)return true;
	bool equal2flag = true;
	for (int i = 0; i < SIZE_X; i++)
	{
		for (int j = 0; j < SIZE_Y; j++)
		{
			if (c[SIZE_X-i-1][j] / 100 != a->c[i][j] / 100)
			{
				equal2flag = false;
				break;
			}
		}
		if (!equal2flag)break;
	}
	return equal2flag;
}

int HakoiriMusume::min(int a, int b)
{
	return a < b ? a : b; 
}
int HakoiriMusume::distance(int x1, int y1, int x2, int y2) 
{
	return abs(x1 - x2) + abs(y1 - y2); 
}

int HakoiriMusume::h() const
{
	int h = 0;
	h += distance(musume_x, musume_y, GOAL_X, GOAL_Y);
	if (hmode == 1) 
	{
		return h * 20;
	}
	int movecost = INT16_MAX;
	if (musume_x > GOAL_X) 
	{
		movecost = min(distance(space_x[0], space_y[0], musume_x - 1, musume_y)
			+ distance(space_x[1], space_y[1], musume_x - 1, musume_y + 1)
			, distance(space_x[1], space_y[1], musume_x - 1, musume_y)
			+ distance(space_x[0], space_y[0], musume_x - 1, musume_y + 1));
	}
	if (musume_x < GOAL_X)
	{
		movecost =min(movecost, min(distance(space_x[0], space_y[0], musume_x + 2, musume_y)
			+ distance(space_x[1], space_y[1], musume_x +2 , musume_y + 1)
			, distance(space_x[1], space_y[1], musume_x +2, musume_y)
			+ distance(space_x[0], space_y[0], musume_x + 2, musume_y + 1)));
	}
	if (musume_y < GOAL_Y)
	{
		movecost = min(movecost, min(distance(space_x[0], space_y[0], musume_x, musume_y + 2)
			+ distance(space_x[1], space_y[1], musume_x + 1, musume_y + 2)
			, distance(space_x[1], space_y[1], musume_x, musume_y + 2)
			+ distance(space_x[0], space_y[0], musume_x + 1, musume_y + 2)));
	}
	if (musume_y > GOAL_Y)
	{
		movecost = min(movecost, min(distance(space_x[0], space_y[0], musume_x, musume_y -1)
			+ distance(space_x[1], space_y[1], musume_x + 1, musume_y - 1)
			, distance(space_x[1], space_y[1], musume_x, musume_y - 1)
			+ distance(space_x[0], space_y[0], musume_x + 1, musume_y -1)));
	}

	h--;
	h = h * 6;
	h += movecost;
	return h;
}

int HakoiriMusume::hash() const 
{
	int h = 0;
	std::hash<int> hasher;
	for (int i = 0; i < SIZE_X; i++)
	{
		for (int j = 0; j < SIZE_Y; j++)
		{
			h += hasher(c[i][j]) % 758699 * (i * 3 + j * 5);
		}
	}
	return h;
}