#pragma once
#include "NoClueSudoku2.h"
#include <fstream>
// #include <Windows.h>

using namespace std;

int NoClueSudoku2::N;
int NoClueSudoku2::GROUP_NUM;
int NoClueSudoku2::BOX_SUM;
int* NoClueSudoku2::GROUP_SIZE;
int** NoClueSudoku2::GROUP_ID;
bool NoClueSudoku2::init(int Qnum) 
{
	char filename[64];
	sprintf(filename, "NoClueSudokuQ%d.txt", Qnum);
	ifstream ifs(filename);
	if (!ifs.good())
	{
		cout << "ԍ܂B" << endl;
		ifs.close();
		return false;
	}
	ifs >> N;
	ifs >> GROUP_NUM;
	GROUP_ID = new int*[N];
	for (int i = 0; i < N; i++)GROUP_ID[i] = new int[N];
	GROUP_SIZE = new int[GROUP_NUM];
	for (int i = 0; i < N;i++)
	{
		for (int j = 0; j < N; j++) 
		{
			ifs >> GROUP_ID[i][j];
		}
	}
	BOX_SUM = N*N*(N + 1) / 2 / GROUP_NUM;
	for (int i = 0; i < GROUP_NUM; i++) 
	{
		GROUP_SIZE[i] = 0;
	}
	for (int i = 0; i < N; i++)
	{
		for (int j = 0; j < N; j++)
		{
			GROUP_SIZE[GROUP_ID[i][j]]++;
		}
	}
	ifs.close();
	return true;
}

void NoClueSudoku2::clean() 
{
	for (int i = 0; i < N; i++)delete[] GROUP_ID[i];
	delete[] GROUP_ID;
	delete[] GROUP_SIZE;
}

NoClueSudoku2::NoClueSudoku2()
{
	c = new int*[N];
	for (int i = 0; i < N; i++)c[i] = new int[N];
	remNumber = new int[N + 1];
	groupSumRem = new int[GROUP_NUM];
	groupNumRem = new int[GROUP_NUM];
	for (int i = 0; i < N; i++)
	{
		for (int j = 0; j < N; j++)
		{
			c[i][j] = 0;
		}
		remNumber[i+1] = N;
	}
	for (int i = 0; i < GROUP_NUM; i++)
	{
		groupSumRem[i] = BOX_SUM;
		groupNumRem[i] = GROUP_SIZE[i];
	}
}

NoClueSudoku2::NoClueSudoku2(const NoClueSudoku2 * p, const NodeBasePtr parent) :NodeBase(parent)
{
	c = new int*[N];
	for (int i = 0; i < N; i++)c[i] = new int[N];
	remNumber = new int[N + 1];
	groupSumRem = new int[GROUP_NUM];
	groupNumRem = new int[GROUP_NUM];
	for (int i = 0; i < N; i++)
	{
		for (int j = 0; j < N; j++)
		{
			c[i][j] = p->c[i][j];
		}
		remNumber[i+1] = p->remNumber[i+1];
	}
	for (int i = 0; i < GROUP_NUM; i++)
	{
		groupSumRem[i] = p->groupSumRem[i];
		groupNumRem[i] = p->groupNumRem[i];
	}
}

NoClueSudoku2::~NoClueSudoku2() 
{
	for (int i = 0; i < N; i++)delete[] c[i];
	delete[] c;
	delete[] groupNumRem;
	delete[] groupSumRem;
	delete[] remNumber;
}

void NoClueSudoku2::expand(std::list<NodeBasePtr>* nodeList, const NodeBasePtr parent) const
{
	int newi, newj;
	for (int i = 0; i < N*N; i++)
	{
		if (c[i / N][i%N] == 0)
		{
			newi = i / N; newj = i%N;
			break;
		}
	}

	bool* enable = new bool[N + 1];
	judgeNG(enable, newi, newj);
	NoClueSudoku2* expandedNode;
	for (int n = 1; n <= N; n++)
	{
		if (!enable[n])continue; //if (!judge(n, newi, newj))continue;
		expandedNode = new NoClueSudoku2(this, parent);
		expandedNode->update(n, newi, newj);
		if (optimize(expandedNode, newi, newj) == false)
		{
			delete expandedNode;
			continue;
		}
		nodeList->push_back(NodeBasePtr(expandedNode));
	}
}

void NoClueSudoku2::judgeNG(bool* enable,int newi,int newj) const
{
	for (int n = 1; n <= N; n++) enable[n] = true;
	
	for (int i = 0; i < N; i++) enable[c[i][newj]] = false;
	
	for (int j = 0; j < N; j++)	enable[c[newi][j]] = false;
	
	int min = groupSumRem[GROUP_ID[newi][newj]] - (groupNumRem[GROUP_ID[newi][newj]] - 1)*N;
	int max = groupSumRem[GROUP_ID[newi][newj]] - (groupNumRem[GROUP_ID[newi][newj]] - 1);
	for (int n = 1; n < min; n++)	enable[n] = false;
	for (int n = max+1; n <= N; n++)	enable[n] = false;
}

bool NoClueSudoku2::optimize(NoClueSudoku2* node,int newi_,int newj_)
{
	NoClueSudoku2* temp = new NoClueSudoku2(node, NodeBasePtr(node));
	stack<pair<int, int>> newijList;
	newijList.push(pair<int, int>(newi_, newj_));
	int newi, newj;
	bool* enableNum=new bool[N+1];

	while (!newijList.empty()) 
	{
		newi = newijList.top().first;
		newj = newijList.top().second;
		newijList.pop();

		//XVsɂčl
		for (int n = 1; n <= N; n++) enableNum[n] = true;
		for (int j = 0; j < N; j++)
		{
			enableNum[node->c[newi][j]] = false;
		}
		for (int n = 1; n <= N; n++)
		{
			if (!enableNum[n])continue;
			int lastpos = -1;
			for (int j = 0; j < N; j++)
			{
				if (node->c[newi][j]>0)continue;
				for (int i = 0; i < N; i++)
				{
					if (node->c[i][j] == n) break;
					if (i == N - 1)
					{
						if (lastpos == -1)lastpos = j;
						else lastpos = -2;
					}
				}
			}
			if (lastpos >= 0)
			{
				if (!node->judge(n, newi, lastpos)) { delete[] enableNum; return false; }
				node->update(n, newi, lastpos);
				newijList.push(pair<int, int>(newi, lastpos));
			}
		}
		//XVɂčl
		for (int n = 1; n <= N; n++) enableNum[n] = true;
		for (int i = 0; i < N; i++) 
		{
			enableNum[node->c[i][newj]] = false;
		}
		for (int n = 1; n <= N; n++) 
		{
			if (!enableNum[n])continue;
			int lastpos = -1;
			for (int i = 0; i < N; i++) 
			{
				if (node->c[i][newj]>0)continue;
				for (int j = 0; j < N; j++) 
				{
					if (node->c[i][j] == n) break;
					if (j == N - 1) 
					{
						if (lastpos == -1)lastpos = i;
						else lastpos = -2;
					}
				}
			}
			if (lastpos >= 0) 
			{
				if (!node->judge(n, lastpos, newj)) { delete[] enableNum; return false; }
				node->update(n, lastpos, newj);
				newijList.push(pair<int, int>(lastpos, newj));
			}
		}
		//XVO[vɂčl
		int newg = GROUP_ID[newi][newj];
		if (node->groupNumRem[newg] == 1)
		{
			for (int i = 0; i < N*N; i++) 
			{
				if (GROUP_ID[i/N][i%N]== newg&&node->c[i/N][i%N]==0) 
				{
					if (!node->judge(node->groupSumRem[newg], i / N, i%N)) { delete[] enableNum; return false; }
					node->update(node->groupSumRem[newg], i / N, i%N);
					newijList.push(pair<int, int>(i / N, i%N));
					break;
				}
			}
		}
		//XVɂčl
		int newn = node->c[newi][newj];
		if (node->remNumber[newn] == 1)
		{
			int posi = -1, posj = -1;
			for (int i = 0; i < N; i++) 
			{
				for (int j = 0; j < N; j++) 
				{
					if (node->c[i][j] == newn)break;
					if (j == N - 1)posi = i;
				}
				if (posi != -1)break;
			}
			for (int j = 0; j < N; j++)
			{
				for (int i = 0; i < N; i++)
				{
					if (node->c[i][j] == newn)break;
					if (i == N - 1)posj = j;
				}
				if (posj != -1)break;
			}
			if (!node->judge(newn, posi, posj)) { delete[] enableNum; return false; }
			node->update(newn,posi,posj);
			newijList.push(pair<int, int>(posi, posj));
		}
		if (!newijList.empty()) 
		{
			newg = 0;
		}
		
		if (hmode == 0)break;
	}

	delete[] enableNum;
	return true;
}

bool NoClueSudoku2::judge(int n, int newi, int newj) const
{
	for (int i = 0; i < N; i++)if (c[i][newj] == n)
	{
		return false;
	}
	for (int j = 0; j < N; j++)if (c[newi][j] == n)
	{
		return false;
	}
	int min = groupSumRem[GROUP_ID[newi][newj]] - (groupNumRem[GROUP_ID[newi][newj]] - 1)*N;
	int max = groupSumRem[GROUP_ID[newi][newj]] - (groupNumRem[GROUP_ID[newi][newj]] - 1);
	if (n < min || max < n)
	{
		return false;
	}

	return true;
}

void NoClueSudoku2::update(int n, int newi, int newj) 
{
	c[newi][newj] = n;
	groupSumRem[GROUP_ID[newi][newj]] -= n;
	groupNumRem[GROUP_ID[newi][newj]]--;
	remNumber[n]--;
}

bool NoClueSudoku2::finished() const
{
	for (int i = 1; i <= N; i++) 
	{
		if (remNumber[i] > 0)return false;
	}
	return true;
}

void NoClueSudoku2::print() const
{
	for (int i = 0; i < N; i++)
	{
		cout << "----";
	}
	cout << endl;
	for (int i = 0; i < N; i++)
	{
		for (int j = 0; j < N; j++)
		{
			cout << c[i][j];
			if (j >= N - 1 || GROUP_ID[i][j] != GROUP_ID[i][j + 1]) cout << " | ";
			else cout << "   ";
		}
		cout << endl;
		if (i >= N - 1)continue;
		for (int j = 0; j < N; j++)
		{
			if (j == 0)
			{
				if (GROUP_ID[i + 1][j] != GROUP_ID[i][j])cout << "---";
				else cout << "   ";
			}
			else
			{
				if (GROUP_ID[i + 1][j] != GROUP_ID[i][j])cout << "----";
				else cout << "    ";
			}
		}
		cout << endl;
	}
	for (int i = 0; i < N; i++)
	{
		cout << "----";
	}
	cout << endl;
}

bool NoClueSudoku2::equalTo(const NodeBasePtr ptr) const
{
	shared_ptr<NoClueSudoku2> a = dynamic_pointer_cast<NoClueSudoku2>(ptr);
	for (int i = 0; i < N; i++)
	{
		for (int j = 0; j < N; j++)
		{
			if (c[i][j] != a->c[i][j])return false;
		}
	}
	return true;
}

int NoClueSudoku2::h() const
{
	int rem = 0;
	for (int i = 1; i <= N; i++) 
	{
		rem += remNumber[i];
	}
	if (hmode == 0)
	{
		return (rem + 4) / 5;
	}

	int hint_number = 0, hint_row = 0, hint_collum = 0, hint_group = 0;
	for (int i = 1; i <= N; i++) 
	{
		if (remNumber[i] == 2)hint_number++;
	}
	for (int i = 0; i < N; i++) 
	{
		int count=0;
		for (int j = 0; j < N; j++) 
		{
			if (c[i][j]==0)count++;
		}
		if (count == 2)hint_collum++;
	}
	for (int j = 0; j < N; j++)
	{
		int count = 0;
		for (int i = 0; i < N; i++)
		{
			if (c[i][j]==0)count++;
		}
		if (count == 2)hint_collum++;
	}
	for (int i = 0; i < GROUP_NUM; i++) 
	{
		if (groupNumRem[i] == 2)hint_group++;
	}

	int temp = rem - hint_collum - hint_row - hint_number - hint_group;
	if (temp > 1)return temp;
	else return  1;
}

int NoClueSudoku2::hash() const
{
	int h = 0;
	std::hash<int> hasher;
	for (int i = 0; i < N; i++)
	{
		for (int j = 0; j < N; j++)
		{
			h += hasher(c[i][j] + i * 3 + j * 11) % 758699 * (5 * i + 7 * j);
			h %= 9999991;
		}
	}
	for (int i = 0; i < GROUP_NUM; i++)
	{
		h += hasher(groupSumRem[i]) % 3266117;
		h += hasher(groupNumRem[i]) % 295201;
	}
	return h;
}