#include "PuzzleSearch.h"
#include <algorithm>

using namespace std;

SearchResult Search(NodeBasePtr start, eSearchMode mode)
{
	SearchResult result;
	NodeWaitList *openNodes;
	NodeHashList closeNodes;
	NodeBasePtr nowNode, tempNode;
	list<NodeBasePtr>* expandedNodes = new list<NodeBasePtr>();

	switch (mode) 
	{
	case SEARCH_DFS:openNodes = new NodeListDFS(); break;
	case SEARCH_DFS2:openNodes = new NodeListDFS2(); break;
	case SEARCH_BFS:openNodes = new NodeListBFS(); break;
	case SEARCH_ASTAR:
		openNodes = new NodeListASTAR();
		NodeBase::hmode = 0;
		break;
	case SEARCH_ASTAR2:
		openNodes = new NodeListASTAR();
		NodeBase::hmode = 1;
		break;
	default:cout << "There isn't such a mode" << endl; return result;
	}

	openNodes->push(start);
	while (!openNodes->empty())
	{
		nowNode = openNodes->pop();
		nowNode->expand(expandedNodes, nowNode);
		while (!expandedNodes->empty())
		{
			tempNode = expandedNodes->front();
			expandedNodes->pop_front();
			result.count++;
			if (result.count % 500 == 0)
			{
				cout << "count=" << result.count;
				cout << " depth=" << tempNode->depth;
				cout << " CLOSE_size=" << closeNodes.size();
				cout << " CLOSE_maxLength=" << closeNodes.maxLength() << endl;
			}
			if (tempNode->finished())
			{
				delete expandedNodes;
				result.goal = tempNode;
				return result;
			}			

			if (closeNodes.contain(tempNode))
			{
				continue;
			}

			openNodes->push(tempNode);
			closeNodes.push(tempNode);
		}
	}
	delete expandedNodes;
	return result;
}

NodeBasePtr NodeListDFS::pop()
{
	NodeBasePtr  node = mStack.top();
	mStack.pop();
	return node;
}
void NodeListDFS::push(const NodeBasePtr node) 
{
	mStack.push(node);
}

NodeBasePtr NodeListBFS::pop()
{
	NodeBasePtr node = mQueue.front();
	mQueue.pop();
	return node;
}
void NodeListBFS::push(const NodeBasePtr node)
{
	mQueue.push(node);
}

NodeBasePtr NodeListASTAR::pop()
{
	pop_heap(mVector.begin(), mVector.end(), MY_COMP_NODE());
	NodeBasePtr node = mVector.back();
	mVector.pop_back();
	return node;
}
void NodeListASTAR::push(const NodeBasePtr node)
{
	mVector.push_back(node);
	push_heap(mVector.begin(), mVector.end(), MY_COMP_NODE());
}

NodeBasePtr NodeListDFS2::pop()
{
	if (empty())return nullptr;
	NodeBasePtr node;
	while (!mList.empty())
	{
		node = mList.front();
		mList.pop_front();
		if (node->depth <= phase*D) return node;
		else mDeepNodes.push_back(node);
	}
	phase += 1;
	mList.splice(mList.begin(), mDeepNodes);
	return pop();
}
void NodeListDFS2::push(const NodeBasePtr node)
{
	if(node->depth > phase*D) mDeepNodes.push_back(node);
	mList.push_front(node);
}

void NodeHashList::push(const NodeBasePtr node)
{
	int hash = node->hash() % HASH_M;
	hashLists[hash].push_back(node);
	mSize++;
	if ((int)hashLists[hash].size() > mMaxLength)mMaxLength = (int)hashLists[hash].size();
}
bool NodeHashList::contain(const NodeBasePtr node) const
{
	int hash = node->hash() % HASH_M;
	for (list<NodeBasePtr>::const_iterator it = hashLists[hash].begin(); it != hashLists[hash].end(); ++it)
	{
		if (node->equalTo((*it)))
		{
			return true;
		}
	}
	return false;
}