#pragma once
#include <list>
#include <queue>
#include <stack>
#include <vector>
#include "NodeBase.h"

enum eSearchMode
{
	SEARCH_DFS,
	SEARCH_DFS2,
	SEARCH_BFS,
	SEARCH_ASTAR,
	SEARCH_ASTAR2,
};

class NodeWaitList
{
public:
	virtual NodeBasePtr pop()=0;
	virtual void push(const NodeBasePtr)=0;
	virtual bool empty() const =0;
};
class NodeListDFS :public NodeWaitList 
{
	std::stack<NodeBasePtr> mStack;
	NodeBasePtr pop();
	void push(const NodeBasePtr);
	bool empty() const { return mStack.empty(); }
};
class NodeListBFS :public NodeWaitList
{
	std::queue<NodeBasePtr> mQueue;
	NodeBasePtr pop();
	void push(const NodeBasePtr);
	bool empty() const { return mQueue.empty(); }
};
class NodeListASTAR :public NodeWaitList 
{
	std::vector<NodeBasePtr> mVector;
	NodeBasePtr pop();
	void push(const NodeBasePtr);
	bool empty() const { return mVector.empty(); }
	class MY_COMP_NODE {
	public:
		bool operator() (const NodeBasePtr left, const NodeBasePtr right) const
		{
			return left->f() > right->f(); //flق傫
		}
	};
};
class NodeListDFS2 :public NodeWaitList 
{
	static const int D = 10;
	int phase;

	std::list<NodeBasePtr> mList;
	std::list<NodeBasePtr> mDeepNodes;

	NodeBasePtr pop();
	void push(const NodeBasePtr);
	bool empty() const { return mList.empty()&&mDeepNodes.empty(); }
public:
	NodeListDFS2() { phase = 1; }
};

class NodeHashList 
{
	static const int HASH_M =16383;
	std::list<NodeBasePtr> hashLists[HASH_M];
	int mSize;
	int mMaxLength;
public:
	NodeHashList() { mSize = 0; mMaxLength = 0; }
	void push(const NodeBasePtr);
	bool contain(const NodeBasePtr) const;
	int size() const { return mSize; }
	int maxLength() const { return mMaxLength; }
};

struct SearchResult
{
	int count;
	NodeBasePtr goal;
	SearchResult() { count = 0; goal = nullptr; }
};

SearchResult Search(NodeBasePtr start, eSearchMode mode);