#include "SlidePuzzle.h"
#include <iostream>
#include <fstream>

using namespace std;

int SlidePuzzle::SIZE_X;
int SlidePuzzle::SIZE_Y;

SlidePuzzle::SlidePuzzle(int Qnum) :NodeBase()
{
	char filename[64];
	sprintf(filename, "SlidePuzzleQ%d.txt", Qnum);
	ifstream ifs(filename);
	ifs >> SIZE_X;
	ifs >> SIZE_Y;
	c = new int*[SIZE_X];
	for (int i = 0; i < SIZE_X; i++)c[i] = new int[SIZE_Y];
	for (int i = 0; i < SIZE_Y; i++)
	{
		for (int j = 0; j < SIZE_X; j++)
		{
			ifs >> c[j][i];
		}
	}
	ifs.close();
}

SlidePuzzle::SlidePuzzle(const SlidePuzzle *p,const NodeBasePtr parent) :NodeBase(parent)
{
	c = new int*[SIZE_X];
	for (int i = 0; i < SIZE_X; i++)c[i] = new int[SIZE_Y];
	for (int i = 0; i < SIZE_X; i++) {
		for (int j = 0; j < SIZE_Y; j++) {
			c[i][j] = p->c[i][j];
		}
	}
}

SlidePuzzle::~SlidePuzzle() 
{
	for (int i = 0; i < SIZE_X; i++)delete[] c[i];
	delete[] c;
}

void SlidePuzzle::expand(list<NodeBasePtr>* nodeList,const NodeBasePtr parent) const
{
	int x0=-1,y0=-1;
	for (int i = 0; i < SIZE_X; i++) 
	{
		for (int j = 0; j < SIZE_Y; j++)
		{
			if (c[i][j] == 0) 
			{
				x0 = i;
				y0 = j;
				break;
			}
		}
		if (x0 != -1)break;
	}
	SlidePuzzle* expandedNode;
	if (x0 > 0) 
	{
		expandedNode = new SlidePuzzle(this, parent);
		expandedNode->c[x0][y0] = c[x0 - 1][y0];
		expandedNode->c[x0 - 1][y0] = 0;
		nodeList->push_back(NodeBasePtr(expandedNode));
	}
	if (x0 < SIZE_X-1)
	{
		expandedNode = new SlidePuzzle(this, parent);
		expandedNode->c[x0][y0] = c[x0 + 1][y0];
		expandedNode->c[x0 + 1][y0] = 0;
		nodeList->push_back(NodeBasePtr(expandedNode));
	}
	if (y0 > 0)
	{
		expandedNode = new SlidePuzzle(this, parent);
		expandedNode->c[x0][y0] = c[x0][y0-1];
		expandedNode->c[x0][y0-1] = 0;
		nodeList->push_back(NodeBasePtr(expandedNode));
	}
	if (y0 < SIZE_Y-1)
	{
		expandedNode = new SlidePuzzle(this, parent);
		expandedNode->c[x0][y0] = c[x0][y0 + 1];
		expandedNode->c[x0][y0 + 1] = 0;
		nodeList->push_back(NodeBasePtr(expandedNode));
	}

}

bool SlidePuzzle::finished() const
{
	int count = 1;
	for (int i = 0; i < SIZE_Y; i++)
	{
		for (int j = 0; j < SIZE_X; j++)
		{
			if (c[j][i] != count)return false;
			count++;
			if (count >= SIZE_X*SIZE_Y)return true;
		}
	}
	return true;
}

void SlidePuzzle::print() const
{
	cout << "-------------" << endl;
	for (int i = 0; i < SIZE_Y; i++) 
	{
		for (int j = 0; j < SIZE_X; j++) 
		{
			if (c[j][i] < 10)cout << c[j][i];
			else {
				char letter = c[j][i] + 97 - 10;
				cout << letter;
			}
			cout << " ";
		}
		cout << endl;
	}
	cout << "-------------" << endl;
}

bool SlidePuzzle::equalTo(const NodeBasePtr ptr) const
{
	shared_ptr<SlidePuzzle> a = dynamic_pointer_cast<SlidePuzzle>(ptr);
	for (int i = 0; i < SIZE_X; i++)
	{
		for (int j = 0; j < SIZE_Y; j++)
		{
			if (c[i][j] != a->c[i][j])return false;
		}
	}
	return true;
}

int SlidePuzzle::h() const
{
	int score = 0;
	for (int i = 0; i < SIZE_X; i++)
	{
		for (int j = 0; j < SIZE_Y; j++)
		{
			score += abs(j - (c[i][j] - 1) / SIZE_X);
			if (c[i][j] == 0) continue;
			else score += abs(i - (c[i][j] - 1) % SIZE_X);
		}
	}
	return score;
}

int SlidePuzzle::hash() const 
{
	int h = 0;
	std::hash<int> hasher;
	for (int i = 0; i < SIZE_X; i++)
	{
		for (int j = 0; j < SIZE_Y; j++)
		{
			h += hasher(c[i][j]) % 758699 * (i * 3 + j * 5);
		}
	}
	return h;
}

